﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  public abstract class KernelBase : ParameterizedNamedItem, ICovarianceFunction {

    #region Parameternames
    private const string DistanceParameterName = "Distance";
    protected const string BetaParameterName = "Beta";
    #endregion
    #region Parameterproperties
    public ValueParameter<IDistance> DistanceParameter {
      get { return Parameters[DistanceParameterName] as ValueParameter<IDistance>; }
    }

    public IFixedValueParameter<DoubleValue> BetaParameter {
      get { return Parameters[BetaParameterName] as FixedValueParameter<DoubleValue>; }
    }

    #endregion
    #region Properties
    public IDistance Distance {
      get { return DistanceParameter.Value; }
    }

    public double Beta {
      get { return BetaParameter.Value.Value; }
    }

    #endregion

    [StorableConstructor]
    protected KernelBase(bool deserializing) : base(deserializing) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() { }

    protected KernelBase(KernelBase original, Cloner cloner)
      : base(original, cloner) { }

    protected KernelBase() {
      Parameters.Add(new ValueParameter<IDistance>(DistanceParameterName, "The distance function used for kernel calculation"));
      DistanceParameter.Value = new EuclideanDistance();
    }

    public double Get(object a, object b) {
      return Get(Distance.Get(a, b));
    }

    protected abstract double Get(double norm);

    public int GetNumberOfParameters(int numberOfVariables) {
      return 1;
    }

    public void SetParameter(double[] p) {
      if (p != null && p.Length == 1) BetaParameter.Value.Value = p[0];
    }

    public ParameterizedCovarianceFunction GetParameterizedCovarianceFunction(double[] p, int[] columnIndices) {
      if (p == null || p.Length != 1) throw new ArgumentException("Illegal parametrization");
      var myClone = (KernelBase)Clone(new Cloner());
      myClone.BetaParameter.Value.Value = p[0];
      var cov = new ParameterizedCovarianceFunction {
        Covariance = (x, i, j) => myClone.Get(GetNorm(x, x, i, j, columnIndices)),
        CrossCovariance = (x, xt, i, j) => myClone.Get(GetNorm(x, xt, i, j, columnIndices)),
        CovarianceGradient = (x, i, j) => new List<double> { myClone.GetGradient(GetNorm(x, x, i, j, columnIndices)) }
      };
      return cov;
    }

    protected abstract double GetGradient(double norm);

    protected double GetNorm(double[,] x, double[,] xt, int i, int j, int[] columnIndices) {
      var dist = Distance as IDistance<IEnumerable<double>>;
      if (dist == null) throw new ArgumentException("The distance needs to apply to double vectors");
      var r1 = new IndexedEnumerable(x, i, columnIndices);
      var r2 = new IndexedEnumerable(xt, j, columnIndices);
      return dist.Get(r1, r2);
    }
    internal class IndexedEnumerable : IEnumerable<double> {
      private readonly double[,] data;
      private readonly int row;
      private readonly int[] columnIndices;

      public IndexedEnumerable(double[,] data, int row, int[] columnIndices) {
        this.data = data;
        this.row = row;
        this.columnIndices = columnIndices;
      }

      public IEnumerator<double> GetEnumerator() {
        return new IndexedEnumerator(data, row, columnIndices);
      }

      IEnumerator IEnumerable.GetEnumerator() {
        return new IndexedEnumerator(data, row, columnIndices);
      }
    }
    internal class IndexedEnumerator : IEnumerator<double> {
      private readonly IEnumerator<int> column;
      private readonly double[,] data;
      private readonly int row;

      public IndexedEnumerator(double[,] data, int row, int[] columnIndices) {
        this.data = data;
        this.row = row;
        column = ((IEnumerable<int>)columnIndices).GetEnumerator();
      }

      public double Current {
        get { return data[row, column.Current]; }
      }

      object IEnumerator.Current {
        get {
          return data[row, column.Current];
        }
      }

      public void Dispose() { }

      public bool MoveNext() {
        return column.MoveNext();
      }

      public void Reset() {
        column.Reset();
      }
    }
  }
}