#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  /// <summary>
  /// Linear regression data analysis algorithm.
  /// </summary>
  [Item("Radial Basis Function Regression (RBF-R)", "Radial basis function regression data analysis algorithm.")]
  [Creatable(CreatableAttribute.Categories.DataAnalysisRegression, Priority = 100)]
  [StorableClass]
  public sealed class RadialBasisRegression : BasicAlgorithm {
    private const string RBFRegressionSolutionResultName = "RBF regression solution";

    public override bool SupportsPause {
      get { return false; }
    }
    public override Type ProblemType {
      get { return typeof(IRegressionProblem); }
    }
    public new IRegressionProblem Problem {
      get { return (IRegressionProblem)base.Problem; }
      set { base.Problem = value; }
    }

    #region parameter names
    private const string KernelParameterName = "Kernel";
    private const string ScaleInputVariablesParameterName = "ScaleInputVariables";
    #endregion

    #region parameter properties
    public ValueParameter<ICovarianceFunction> KernelParameter {
      get { return (ValueParameter<ICovarianceFunction>)Parameters[KernelParameterName]; }
    }

    public IFixedValueParameter<BoolValue> ScaleInputVariablesParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[ScaleInputVariablesParameterName]; }
    }
    #endregion

    #region properties
    public ICovarianceFunction Kernel {
      get { return KernelParameter.Value; }
    }

    public bool ScaleInputVariables {
      get { return ScaleInputVariablesParameter.Value.Value; }
      set { ScaleInputVariablesParameter.Value.Value = value; }
    }

    #endregion

    [StorableConstructor]
    private RadialBasisRegression(bool deserializing) : base(deserializing) { }
    private RadialBasisRegression(RadialBasisRegression original, Cloner cloner)
      : base(original, cloner) {
    }
    public RadialBasisRegression() {
      Problem = new RegressionProblem();
      Parameters.Add(new ValueParameter<ICovarianceFunction>(KernelParameterName, "The radial basis function"));
      Parameters.Add(new FixedValueParameter<BoolValue>(ScaleInputVariablesParameterName, "Set to true if the input variables should be scaled to the interval [0..1]", new BoolValue(true)));
      var kernel = new GaussianKernel();
      KernelParameter.Value = kernel;
    }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RadialBasisRegression(this, cloner);
    }

    protected override void Run(CancellationToken cancellationToken) {
      double loocvrmse, rmsError;
      var solution = CreateRadialBasisRegressionSolution(Problem.ProblemData, Kernel, ScaleInputVariables, out loocvrmse, out rmsError);
      Results.Add(new Result(RBFRegressionSolutionResultName, "The RBF regression solution.", solution));
      Results.Add(new Result("LOOCVRMSE", "The root mean squared error of a leave-one-out-cross-validation on the training set", new DoubleValue(loocvrmse)));
      Results.Add(new Result("RMSE (test)", "The root mean squared error of the solution on the test set.", new DoubleValue(rmsError)));
    }

    public static IRegressionSolution CreateRadialBasisRegressionSolution(IRegressionProblemData problemData, ICovarianceFunction kernel, bool scaleInputs, out double loocvRmsError, out double rmsError) {
      var model = new RadialBasisFunctionModel(problemData.Dataset, problemData.TargetVariable, problemData.AllowedInputVariables, problemData.TrainingIndices, scaleInputs, kernel);
      loocvRmsError = model.LeaveOneOutCrossValidationRootMeanSquaredError();
      rmsError = Math.Sqrt(model.GetEstimatedValues(problemData.Dataset, problemData.TestIndices)
        .Zip(problemData.TargetVariableTestValues, (a, b) => (a - b) * (a - b))
        .Average());
      var solution = model.CreateRegressionSolution((IRegressionProblemData)problemData.Clone());
      solution.Model.Name = "RBF Regression Model";
      solution.Name = "RBF Regression Solution";
      return solution;
    }
  }
}
